#include "BMPRW.h"

BMPRW::BMPRW()
{}

BMPRW::~BMPRW()
{
delete image;
}


BMPRW::BMPRW(char *sName,char *dName)
{
  image = new unsigned char[NUM_PIXELS];
  sourcename = new char[strlen(sName)];
  destinationname = new char[strlen(dName)];
  strcpy(sourcename,sName);
  strcpy(destinationname,dName);
} 


bool
BMPRW::Read()
{
  bool val = ReadHeader() && CheckHeader() && ReadPalette() && ReadImageData();
  CloseSourceFile();
  return val;
}


bool
BMPRW::Write()
{
  bool val = WriteHeader() && WritePalette() && WriteImageData();
  CloseDestFile();
  return val;
}

 
bool
BMPRW::ReadHeader()
{
  sourcefile = fopen(sourcename,"r");
  if(sourcefile==NULL){
    printf("Error opening file: %s\n%s\n",sourcename,strerror(errno));
    return false;
  }

  printf("Reading file header. ");
  returnval = fread(header,1,BMP_HEADERSIZE,sourcefile);
  printf("Read %d bytes.\n",returnval);
  if(returnval != BMP_HEADERSIZE){
    printf("Read error. %s\n",strerror(errno));
    return false;
  }
  return true;
}

bool
BMPRW::CheckHeader()
{
  if(*(unsigned int*)(header+BMP_OFFSET_WIDTH)!=176){
    printf("BMP width must be %d   Was: %d\n",IMAGE_WIDTH,*(unsigned int*)(header+BMP_OFFSET_WIDTH));
    return false;
  }

  if(*(unsigned int*)(header+BMP_OFFSET_HEIGHT)!=144){
    printf("BMP height must be %d   Was: %d\n",IMAGE_WIDTH,*(unsigned int*)(header+BMP_OFFSET_HEIGHT));
    return false;
  }

  if(*(unsigned int*)(header+BMP_OFFSET_COMPRESSION)!=0){
    printf("BMP is compressed. Only uncompressed image supported.\n");
    return false;
  }

  if(*(unsigned short*)(header+BMP_OFFSET_BPP)!=8){
    printf("BMP must be %d bpp. Was: %u bpp\n",BMP_NUMBPP,*(unsigned short*)(header+BMP_OFFSET_BPP));
    return false;
  }

  if(*(unsigned int*)(header+BMP_OFFSET_NUMCOLOURS)!=256){  // 0 means 256 entries
    printf("BMP must have %d palette entries. Had: %u entries\n",BMP_NUMCOLOURS,*(unsigned int*)(header+BMP_OFFSET_NUMCOLOURS));
    return false;
  }

  printf("Header seems ok. Assuming grayscale palette.\n");

  return true;
}

bool
BMPRW::ReadPalette()
{
  printf("Reading palette. ");
  returnval = fread(palette,sizeof(char)*4,BMP_NUMCOLOURS,sourcefile);
  printf("Read %d bytes.\n",returnval*4);
  if(returnval != BMP_NUMCOLOURS){
    printf("Read error. %s\n",strerror(errno));
    return false;
  }
  return true;
}

bool
BMPRW::ReadImageData()
{
  printf("Reading image data. ");
  returnval = fread(image,sizeof(char),NUM_PIXELS,sourcefile);
  printf("Read %d bytes.\n",returnval);
  if(returnval != NUM_PIXELS){
    printf("Read error. %s\n",strerror(errno));
    return false;
  }

  return true;
}


bool
BMPRW::WriteHeader()
{
  destinationfile = fopen(destinationname,"w+");
  if(destinationfile == NULL){
    printf("Error creating file: %s\n%s\n",destinationname,strerror(errno));
    return false;
  }
  
  printf("Writing header. ");
  returnval = fwrite(header,1,BMP_HEADERSIZE,destinationfile);
  if(returnval != BMP_HEADERSIZE){
    printf("Write error. %s\n",strerror(errno));
    return false;
  }
  printf("Wrote %d bytes.\n",returnval);
  return true;
}

bool
BMPRW::WritePalette()
{
  printf("Writing palette. ");
  returnval = fwrite(palette,sizeof(char)*4,BMP_NUMCOLOURS,destinationfile);
  if(returnval != BMP_NUMCOLOURS){
    printf("Write error. %s\n",strerror(errno));
    return false;
  }
  printf("Wrote %d bytes.\n",BMP_NUMCOLOURS*4);
  return true;
}

bool
BMPRW::WriteImageData()
{
  printf("Writing image data. ");
  returnval = fwrite(image,sizeof(char),NUM_PIXELS,destinationfile);
  if(returnval != NUM_PIXELS){
    printf("Write error. %s\n",strerror(errno));
    return false;
  }
  printf("Wrote %d bytes.\n",returnval);
  return true;
}

void
BMPRW::CloseDestFile()
{
  if(destinationfile != NULL){
    printf("Closing destination file.\n");
    fclose(destinationfile);
  }
}


void
BMPRW::CloseSourceFile()
{
  if(sourcefile != NULL){
    printf("Closing source file.\n");
    fclose(sourcefile);
  }
}


unsigned char*
BMPRW::GetImageData()
{
  return image;
}

void
BMPRW::SetImageData(unsigned char *data)
{
	if(image != NULL)
		delete image;
	image = data;
}
