#include <OPENR/OSyslog.h>
#include <OPENR/core_macro.h>
#include <OPENR/OPENRAPI.h>
#include <math.h>
#include "NiloImage.h"
#include "RecognizeBall.h"
#include "Constants.h"

NiloImage::NiloImage() : fbkID(oprimitiveID_UNDEF)
{
	recognizeBall = new RecognizeBall(NULL);
}

NiloImage::~NiloImage()
{
	delete recognizeBall;
}

	OStatus
NiloImage::DoInit(const OSystemEvent& event)
{
	NEW_ALL_SUBJECT_AND_OBSERVER;
	REGISTER_ALL_ENTRY;
	SET_ALL_READY_AND_NOTIFY_ENTRY;

	OpenPrimitive();
}

	OStatus
NiloImage::DoStart(const OSystemEvent& event)
{
	ENABLE_ALL_SUBJECT;
	ASSERT_READY_TO_ALL_OBSERVER;
	return oSUCCESS;
}

	OStatus
NiloImage::DoStop(const OSystemEvent& event)
{
	DISABLE_ALL_SUBJECT;
	DEASSERT_READY_TO_ALL_OBSERVER;
	return oSUCCESS;
}

	OStatus
NiloImage::DoDestroy(const OSystemEvent& event)
{
	DELETE_ALL_SUBJECT_AND_OBSERVER;
	return oSUCCESS;
}

	void
NiloImage::Notify(const ONotifyEvent& event)
{
	static int counter = 0;

	/* Receive the image vector */
	OFbkImageVectorData* imageVectorData = (OFbkImageVectorData*)event.Data(0);
	printf("Received image vector.\n");
	OFbkImage yImage = OFbkImage(imageVectorData->GetInfo(ofbkimageLAYER_H), 
			imageVectorData->GetData(ofbkimageLAYER_H), ofbkimageBAND_Y);


	/* Copy the image
	*/
	byte *yImageData = new byte[ORIG_IMAGE_SIZE];
	for(int i=0; i<ORIG_IMAGE_HEIGHT; i++)
		for(int j=0; j<ORIG_IMAGE_WIDTH; j++)
			yImageData[i*ORIG_IMAGE_WIDTH+j] = yImage.Pixel(j,i);

	//RUNNING
	recognizeBall->YImage(yImageData);
	int *ball_info = recognizeBall->DetectBall(); //Not running :)
	delete yImageData;
	observer[event.ObsIndex()]->AssertReady(event.SenderID());
}

	void
NiloImage::OpenPrimitive()
{
	OStatus result = OPENR::OpenPrimitive(FBK_LOCATOR, &fbkID);
	if(result != oSUCCESS)
		printf("OpenPrimitive(): Couldn't open camera!");
}
