#include <stdio.h>
#include <math.h>
#include "Table.h"
#include "Constants.h"
#include "RecognizeBall.h"

/////// PRIVATE ///////

/* UTILITIES */

    int
RecognizeBall::Fac(int n)
{
    int result = 1;
    while(n > 0){
        result = result * n;
        n = n - 1;
    }
    return result;
}

    int
RecognizeBall::Combination(int n, int k)
{
    int delta = k - 1;
    int prod = 1;
    while(delta >= 0){
        prod = prod * (n-delta);
        delta--;
    }
    return prod / Fac(k);
    //return Fac(n)/(Fac(k)*Fac(n-k));
}

    void 
RecognizeBall::KillBloblist(bloblist_t *bloblist)
{
    int cntr = 0;
    while(bloblist != NULL){
        bloblist_t *tmp = bloblist->next;
        if(bloblist->blob != NULL){
            delete bloblist->blob;
        }
        delete bloblist;
        bloblist = tmp;
        cntr++;
    }
}

    int
RecognizeBall::IsElement(int element, int array[], int *arraylen)
{
    int len = *arraylen;
    for(int i=0; i<len; i++)
        if(array[i] == element)
            return 1;
    return 0;
}

    int
RecognizeBall::CountBlobs(bloblist_t *bloblist)
{
    bloblist_t *tmp = bloblist;
    int cntr = 0;
    while(tmp != NULL){
        cntr++;
        tmp = tmp->next;
    }
    return cntr;
}


/* ANALYSIS SECTION */

    int
RecognizeBall::CheckPixel(int x, int y)
{
    return !(x<0 || y<0 || x>=SOBEL_IMAGE_WIDTH || 
            y>=SOBEL_IMAGE_HEIGHT ||
            sImage[YX_TO_INDEX[y][x]]==ON_PIXEL ||
            sImage[YX_TO_INDEX[y][x]]==BLOB_PIXEL);
}

    AdHocList*
RecognizeBall::MakeBlob(int offset)
{
    AdHocList *blob = new AdHocList();
    AdHocList *head = new AdHocList();
    const int *coord = INDEX_TO_YX[offset];

    if(CheckPixel(coord[X_INDEX],coord[Y_INDEX])){
        head->Push(offset);
        blob->Push(offset);
        sImage[offset] = BLOB_PIXEL;
    }

    while(head->Size() > 0){
        offset = head->Pop();

        if(blob->Size() > MAX_BLOB_SIZE){ //Rollback action
            while(blob->Size() > 0)
                sImage[blob->Pop()] = OFF_PIXEL;
            while(head->Size() > 0)
                sImage[head->Pop()] = OFF_PIXEL;
            delete head;
            delete blob;
            return NULL;
        }
        coord = INDEX_TO_YX[offset];
        int y = coord[Y_INDEX];
        int x = coord[X_INDEX];
        int index = -1;

        if(CheckPixel(x-1,y)){
            index = offset - 1;
            head->Push(index);
            sImage[index] = BLOB_PIXEL;
            blob->Push(index);
        }
        if(CheckPixel(x,y-1)){
            index = offset - SOBEL_IMAGE_WIDTH;
            head->Push(index);
            sImage[index] = BLOB_PIXEL;
            blob->Push(index);
        }
        if(CheckPixel(x+1,y)){
            index = offset + 1;
            head->Push(index);
            sImage[index] = BLOB_PIXEL;
            blob->Push(index);
        }
        if(CheckPixel(x,y+1)){
            index = offset + SOBEL_IMAGE_WIDTH;
            head->Push(index);
            sImage[index] = BLOB_PIXEL;
            blob->Push(index);
        }
    }
    delete head;
    if(blob->Size() < MIN_BLOB_SIZE){
        while(blob->Size() > 0)
            sImage[blob->Pop()] = OFF_PIXEL;
        delete blob;
        blob = NULL;
    }
    return blob;
}


    int 
RecognizeBall::CheckPerimeter(int offset)
{
    int left = 0;
    int right = 0;
    const int *coord = INDEX_TO_YX[offset];
    int y = coord[Y_INDEX];
    int x = coord[X_INDEX];
    for(int i=0; i<MAX_DISTANCE_TO_PIXEL; i++){
        if(x+i < SOBEL_IMAGE_WIDTH && sImage[offset+i] == ON_PIXEL)
            right = 1;
        if(x-i >= 0 && sImage[offset-i] == ON_PIXEL)
            left = 1;
    }
    return right && left;
}

    int
RecognizeBall::FindPotentialBlob(int offset)
{
    int blackStart = -1;
    int blackStop = -1;
    int half = -1;
    for(int i=1; i<MAX_DISTANCE_TO_PIXEL; i++){
        int index = offset+i*SOBEL_IMAGE_WIDTH;
        if(index >= SOBEL_IMAGE_SIZE)
            return -1;
        byte pixel = sImage[index];
        if(blackStart < 0 && pixel == OFF_PIXEL){
            blackStart = index;
        }
        else if(blackStart >= 0 && blackStop < 0 && pixel == ON_PIXEL){
            blackStop = index - SOBEL_IMAGE_WIDTH; 
            half = blackStart + 
                (((blackStop - blackStart) / 175) / 2) * 175; 
            if(CheckPerimeter(half)){
                return half;
            }
        }	
    }
    return -1;
}

    int
RecognizeBall::FindNextOnPixel(int offset)
{
    for(int i=offset; i<SOBEL_IMAGE_SIZE; i++)
        if(sImage[i] == ON_PIXEL) 
            return i;
    return -1; //no more on pixels in the image
}

    bloblist_t* 
RecognizeBall::FindAllBlobs()
{
    bloblist_t *bloblist = new bloblist_t(NULL, NULL);
    bloblist_t *currblob = bloblist;
    int onPixel = FindNextOnPixel(0);
    while(onPixel >= 0){
        int regionGrowingOffset = FindPotentialBlob(onPixel);
        if(regionGrowingOffset > -1){
            AdHocList *tmpBlob = MakeBlob(regionGrowingOffset);
            if(tmpBlob != NULL){ 
                currblob->next = 
                    new bloblist_t(tmpBlob, NULL);
                currblob = currblob->next;
            }
        }
        onPixel = FindNextOnPixel(onPixel+1);
    }
    currblob = bloblist->next;
    delete bloblist;

    return currblob;
}

    AdHocList*
RecognizeBall::CalcBlobCenter(bloblist_t *bloblist)
{
    bloblist_t *curr = bloblist;
    AdHocList *blobcenter = new AdHocList();
    const int *yx;

    while(curr != NULL){
        AdHocList *blob = curr->blob;
        yx = INDEX_TO_YX[blob->Max()];
        int yMin = yx[Y_INDEX];
        yx = INDEX_TO_YX[blob->Min()];
        int yMax = yx[Y_INDEX];
        int ycenter = yMin + (yMax - yMin)/2;
        int xMin = 1000;
        int xMax = -1000;
        for(int i=0; i<blob->Size(); i++){
            yx = INDEX_TO_YX[blob->Get(i)];
            int x = yx[X_INDEX];

            if(x > xMax)
                xMax = x;
            if(x < xMin)
                xMin = x;
        }
        int xcenter = xMin + (xMax - xMin)/2;
        blobcenter->Push(YX_TO_INDEX[ycenter][xcenter]);
        curr = curr->next;
    }
    return blobcenter;
}

    float
RecognizeBall::CalcDistanceToBall(int dist)
{
    return dist * SCALE_TO_REAL_WORLD;
}

    point_t
RecognizeBall::CalcCenterOfBall(point_t p1, point_t p2, point_t p3)
{
    int Minx = MIN(p1.x, p2.x);
    Minx = (Minx < p3.x) ? Minx : p3.x;
    int Maxx = MAX(p1.x, p2.x);
    Maxx = (Maxx > p3.x) ? Maxx : p3.x;
    int Miny = MIN(p1.y, p2.y);
    Miny = (Miny < p3.y) ? Miny : p3.y;
    int Maxy = MAX(p1.y, p2.y);
    Maxy = (Maxy > p3.y) ? Maxy : p3.y;
    int cx = Minx + (Maxx-Minx)/2;
    int cy = Miny + (Maxy-Miny)/2;
    point_t p;
    p.x=cx;
    p.y=cy;

    return p;
}

    float 
RecognizeBall::CalcAngle(vector_t u, vector_t v)
{
    float dotprod = u.x * v.x + u.y * v.y;
    float ulen = sqrt((double)(u.x*u.x + u.y*u.y));
    float vlen = sqrt((double)(v.x*v.x + v.y*v.y));
    return dotprod/(ulen*vlen);
}

    vector_t
RecognizeBall::CalcVector(point_t p1, point_t p2)
{
    vector_t vector(p2.x-p1.x, p2.y-p1.y); 
    return vector;
}

    int
RecognizeBall::CheckAngles(point_t p1, point_t p2, point_t p3)
{
    vector_t u = CalcVector(p1,p2);
    vector_t v = CalcVector(p1,p3);
    float angle1 = CalcAngle(u,v);
    u = CalcVector(p2,p1);
    v = CalcVector(p2,p3);
    float angle2 = CalcAngle(u,v);
    u = CalcVector(p3,p1);
    v = CalcVector(p3,p2);
    float angle3 = CalcAngle(u,v);

    //# The reason for the inverted comparisment is
    //# Due to the fact that cos(70) < cos(50)
    if(angle1 < MAX_ANGLE || angle1 > MIN_ANGLE)
        return 0;
    if(angle2 < MAX_ANGLE || angle2 > MIN_ANGLE)
        return 0;
    if(angle3 < MAX_ANGLE || angle3 > MIN_ANGLE)
        return 0;
    return 1;
}

    float
RecognizeBall::CalcDistance(point_t p1, point_t p2)
{
    return (float)sqrt(pow((double)(p1.x-p2.x),(double)2) + 
            pow((double)(p1.y-p2.y),(double)2));
}

    int
RecognizeBall::CheckDistances(point_t p1, point_t p2, point_t p3, float *avg)
{
    float dist1 = CalcDistance(p1,p2);
    float dist2 = CalcDistance(p1,p3);
    float dist3 = CalcDistance(p2,p3);
    *avg = (dist1+dist2+dist3) / 3.0;
    if(dist1 > *avg + DISTANCE_TOLERANCE || dist1 < *avg - DISTANCE_TOLERANCE)
        return 0;
    if(dist2 > *avg + DISTANCE_TOLERANCE || dist2 < *avg - DISTANCE_TOLERANCE)
        return 0;
    if(dist3 > *avg + DISTANCE_TOLERANCE || dist3 < *avg - DISTANCE_TOLERANCE)
        return 0;
    return 1;
}

    void
RecognizeBall::MakeTriangles(int array[], int arraylen, triangle_t result[])
{
    int nbrOfCombs = Combination(arraylen,3);
    int i = 0;
    int j = 1;
    int k = 2;
    for(int cntr=0; cntr<nbrOfCombs; cntr++){
        const int *yx1 = INDEX_TO_YX[array[i]];
        const int *yx2 = INDEX_TO_YX[array[j]];
        const int *yx3 = INDEX_TO_YX[array[k]];
        point_t p1(yx1[X_INDEX], yx1[Y_INDEX]);
        point_t p2(yx2[X_INDEX], yx2[Y_INDEX]);
        point_t p3(yx3[X_INDEX], yx3[Y_INDEX]);
        triangle_t triangle(p1, p2, p3);
        result[cntr] = triangle;
        k += 1;
        if(k >= arraylen){
            j += 1;
            k = j + 1;
            if(j + 1 >= arraylen){
                i += 1;
                j = i + 1;
                k = j + 1;
                if(i + 2 >= arraylen){
                    return; 
                }
            }
        }
    }
}

    int
RecognizeBall::AnalyzeImage(AdHocList *blobcenter, ball_t *ball)
{
    int numblobs = blobcenter->Size();
    int *blobarray = new int[numblobs];
    blobcenter->ToArray(blobarray);
    int numcombs = Combination(numblobs,3);
    triangle_t *triangles = new triangle_t[numcombs];
    MakeTriangles(blobarray,numblobs,triangles);
    delete [] blobarray;

    for(int i=0; i<numcombs; i++){
        triangle_t triangle = triangles[i];
        float avg = 0;
        if(CheckDistances(triangle.p1, triangle.p2, triangle.p3, &avg)){
            if(CheckAngles(triangle.p1, triangle.p2, triangle.p3)){
                point_t p = CalcCenterOfBall(triangle.p1, 
                        triangle.p2,triangle.p3);
                ball->x = p.x;
                ball->y = p.y;
                ball->size = avg;// * SCALE_TO_REAL_WORLD; 
                delete triangles;
                return 1;
            }
        }

    }
    delete triangles;
    return 0;
}


/* IMAGE PROCESSING */

    byte*
RecognizeBall::ApplyThreshold(int threshold)
{
    for(int i=0; i<SOBEL_IMAGE_SIZE; i++)
        if(sImage[i] > threshold)
            sImage[i] = ON_PIXEL; //white
        else
            sImage[i] = OFF_PIXEL; //black
    return sImage;
}

    int
RecognizeBall::CalculateThreshold()
{
    int MaxVal = 0;
    int MaxInd = 0;
    for(int i=0; i<NUMBER_OF_GREYS; i++)
        if(MaxVal < histogram[i]){
            MaxVal = histogram[i];
            MaxInd = i;
        }
    int startPoint = MaxInd;
    int endPoint = MaxInd + INTERVAL;
    while(endPoint < NUMBER_OF_GREYS){
        int deltaX = INTERVAL; //endPoint - startPoint
        int deltaY = histogram[endPoint] - histogram[startPoint];
        float k = 0;
        if(deltaX == 0)
            k=deltaY;
        else
            k = (float)deltaY / (float)deltaX; //the slope
        if(k >= K_MIN)
            return startPoint;
        else{
            startPoint += INTERVAL;
            endPoint += INTERVAL;
        }
    }
    return -1; //We are not happy here
}

    int*
RecognizeBall::SmoothenHistogram()
{
    for(int i=0; i<NUMBER_OF_GREYS - SMOOTH_INTERVAL; i++)
        for(int j=0; j<SMOOTH_INTERVAL-1; j++){
            histogram[i]+=histogram[i+j+1];
            //This division is not really needed since
            //the slope will be the same anyway
            histogram[i] = histogram[i] / SMOOTH_INTERVAL;
        }
    return histogram;
}

    int*
RecognizeBall::MakeHistogram()
{
    for(int i=0; i<SOBEL_IMAGE_SIZE; i++)
        histogram[sImage[i]]++;
    return histogram;
}

    void
RecognizeBall::ClearHistogram()
{
    for(int i=0; i<NUMBER_OF_GREYS; i++)
        histogram[i] = 0;
}

    byte* 
RecognizeBall::ApplySobelMask()
{
    int cntr = 0;
    int offset = 0;
    for(int i=0; i<SOBEL_IMAGE_SIZE; i++){
        if(cntr == ORIG_IMAGE_WIDTH-1){
            offset += 1;
            cntr = 0;
        }
        int row1 = i + offset;
        int row2 = i + ORIG_IMAGE_WIDTH + offset;
        double vertical = yImage[row1] - yImage[row2+1];
        double horizontal = yImage[row1+1] - yImage[row2];
        int result = (int)(sqrt(pow(vertical,(double)2) + 
                    pow(horizontal,(double)2)));
        sImage[i] = (byte)result;
        cntr += 1;
    }
}

    void
RecognizeBall::ProcessImage()
{
    ApplySobelMask();
    ClearHistogram();
    MakeHistogram();
    SmoothenHistogram();
    int threshold = CalculateThreshold();
    ApplyThreshold(threshold);  
}


/////// PUBLIC ///////

    int*
RecognizeBall::DetectBall()
{
    ProcessImage();
    bloblist_t *bloblist = FindAllBlobs();
    int nbrOfBlobs = CountBlobs(bloblist);
    if(nbrOfBlobs < 3){
        printf("Found %d blobs ==> Could not detect ball!!\n", 
                nbrOfBlobs);
        KillBloblist(bloblist);
        return NULL;
    }
    AdHocList *blobcenterlist = CalcBlobCenter(bloblist);
    FilterWhitePixels(); //DEBUG
    MarkBlobCenter(blobcenterlist);// DEBUG
    ball_t *ball = new ball_t(-1,-1,-1);
    if(AnalyzeImage(blobcenterlist, ball)){
        printf("detected ball at x=%d, y=%d with a size of %f\n"
                ,ball->x, ball->y, ball->size);
        sImage[YX_TO_INDEX[ball->y][ball->x]] = ON_PIXEL; 
    }else{
        printf("Could not detect ball!!\n");
    }
    delete ball;
    delete blobcenterlist;
    KillBloblist(bloblist);
    return NULL;
}

    byte*
RecognizeBall::PaddedSobelImage(byte psImage[])
{
    for(int i=0; i<SOBEL_IMAGE_HEIGHT; i++){
        for(int j=0; j<SOBEL_IMAGE_WIDTH; j++){
            int index = i*SOBEL_IMAGE_WIDTH+j;
            psImage[index+i] = sImage[index];	
        }
        psImage[i*ORIG_IMAGE_WIDTH+SOBEL_IMAGE_WIDTH] = OFF_PIXEL;
    }
    for(int i=0; i<ORIG_IMAGE_WIDTH; i++)
        psImage[(ORIG_IMAGE_HEIGHT - 1) * ORIG_IMAGE_WIDTH + i] = OFF_PIXEL;
    return psImage;
}

    byte*
RecognizeBall::SobelImage()
{ return this->sImage; }

    byte*
RecognizeBall::YImage()
{ return this->yImage; }

    void 
RecognizeBall::YImage(byte *yImage)
{ this->yImage = yImage; }

RecognizeBall::RecognizeBall(byte *yImage)
{
    this->yImage = yImage;
    this->sImage = new byte[SOBEL_IMAGE_SIZE];
    this->histogram = new int[256];
}

RecognizeBall::~RecognizeBall()
{
    delete [] histogram;
    delete [] sImage;
}


