#ifndef RecognizeBall_H
#define RecognizeBall_H

#include "AdHocList.h"

#define byte unsigned char

//Structs
struct bloblist_t
{
    AdHocList *blob;
    bloblist_t *next;

    bloblist_t()
    {
        blob = NULL;
        next = NULL;
    }
    bloblist_t(AdHocList *bl, bloblist_t *ne)
    {
        blob = bl;
        next = ne;
    }
};

struct vector_t
{
    int x;
    int y;

    vector_t(int newx, int newy)
    {
        x = newx;
        y = newy;
    }

};

struct point_t
{
    int x;
    int y;

    point_t(){}
    point_t(int newx, int newy)
    {
        x = newx;
        y = newy;
    }
};

struct triangle_t
{
    point_t p1;
    point_t p2;
    point_t p3;

    triangle_t(){}
    triangle_t(point_t newp1, point_t newp2, point_t newp3)
    {
        p1 = newp1;
        p2 = newp2;
        p3 = newp3;
    }
};

struct ball_t
{
    int x;
    int y;
    float size;

    ball_t(int newx, int newy, float newsize)
    {
        x = newx;
        y = newy;
        size = newsize;
    }
};

class RecognizeBall
{
    private:
        //Attributes
        byte *yImage; //original y image
        byte *sImage; //y image with sobel filter applied
        int *histogram; //histogram of sImage

        // DEBUGGING 
        void FlushIt(); //For poor wireless console
        int PrintBloblist(bloblist_t *bloblist);
        void PrintHistogram();
        void MarkBlobCenter(AdHocList *blobcenterlist);
        void FilterWhitePixels();

        //utilities
        int IsElement(int element, int array[], int *arraylen);
        void KillBloblist(bloblist_t *bloblist);
        int Fac(int n);
        int Combination(int n, int k);
        int CountBlobs(bloblist_t *bloblist);

        //Image processing
        byte* ApplyThreshold(int threshold);
        int CalculateThreshold();
        int* SmoothenHistogram();
        int* MakeHistogram();
        void ClearHistogram();
        byte* ApplySobelMask();
        void ProcessImage();

        //Image analysis
        float CalcDistanceToBall(int dist);
        point_t CalcCenterOfBall(point_t p1, point_t p2, point_t p3);
        vector_t CalcVector(point_t p1, point_t p2); 
        float CalcAngle(vector_t u, vector_t v);
        int CheckAngles(point_t p1, point_t p2, point_t p3);
        float CalcDistance(point_t p1, point_t p2);
        int CheckDistances(point_t p1, point_t p2, point_t p3, 
                        float *avg);
        void MakeTriangles(int array[], int arraylen, 
                        triangle_t triangles[]);
        int AnalyzeImage(AdHocList *blobcenter, ball_t *ball);
        AdHocList* CalcBlobCenter(bloblist_t *bloblist);
        int CheckPixel(int x, int y);
        AdHocList* MakeBlob(int offset);
        int CheckPerimeter(int offset);
        int FindPotentialBlob(int offset);
        int FindNextOnPixel(int offset);
        bloblist_t* FindAllBlobs();

    public:
        RecognizeBall(byte *yImage);
        ~RecognizeBall();
        int* DetectBall();
        byte* SobelImage(); 
        byte* PaddedSobelImage(byte psImage[]);
        byte* YImage();
        void YImage(byte *yImage);
};

#endif
