import Image
import ImageFilter
import profile
import pstats
from math import *
from lookup import *
from general_convolution import *

#Thresholding
K_MIN = -0.2
K_MAX = 0.02 #not currently used
INTERVAL = 5 

#Histogram
SMOOTH_INTERVAL = 4

#ON pixel
MAX_DISTANCE_TO_PIXEL = 11 #10 seems reasonable for this size
ON_PIXEL = 255
OFF_PIXEL = 0
BLOB_PIXEL = 128
MAX_BLOB_RADIUS = 8
MAX_BLOB_SIZE = MAX_BLOB_RADIUS**2*pi
MIN_BLOB_SIZE = 6
DISTANCE_TOLERANCE = 2
ANGLE_TOLERANCE = 10
OPTIMAL_ANGLE = 60
MIN_ANGLE = cos((OPTIMAL_ANGLE - ANGLE_TOLERANCE)*(pi/180))
MAX_ANGLE = cos((OPTIMAL_ANGLE + ANGLE_TOLERANCE)*(pi/180))
SCALE_TO_REAL_WORLD = 10/100.0 #I have nooo clue about this!! Needs testing

#Image
ORIG_IMAGE_WIDTH = 176
ORIG_IMAGE_HEIGHT = 144
ORIG_IMAGE_SIZE = ORIG_IMAGE_WIDTH * ORIG_IMAGE_HEIGHT
SOBEL_IMAGE_WIDTH = 175
SOBEL_IMAGE_HEIGHT = 143
SOBEL_IMAGE_SIZE = SOBEL_IMAGE_WIDTH * SOBEL_IMAGE_HEIGHT

def sobelYXToIndex(y,x):
    return SOBEL_LOOKUP_YX[y][x]

def sobelIndexToYX(index):
    return SOBEL_LOOKUP_INDEX[index]

def fastSobel(oldIm):
    """Detect edges and lines in an image"""
    newIm = []
    cntr = 0
    offset = 0
    imsize = SOBEL_IMAGE_SIZE
    for i in range(imsize):
        if cntr == ORIG_IMAGE_WIDTH-1:
            offset += 1
            cntr = 0
        row1 = i + offset
        row2 = i + ORIG_IMAGE_WIDTH + offset
        vertical = oldIm[row1] - oldIm[row2+1]
        horizontal = oldIm[row1+1] - oldIm[row2]
        result = pow(vertical**2 + horizontal**2, 0.5)
        newIm.append(result)
        cntr += 1
    return newIm
    
def histogram(image):
    """Create a histogram for an image"""
    histogram = [0 for i in range(256)]
    for i in range(len(image)):
        index = int(image[i])
        histogram[index] += 1
    return histogram

def smooth_histogram(histogram):
    """Smooth out peaks and valleys"""
    for i in range(len(histogram)-SMOOTH_INTERVAL):
        for j in range(SMOOTH_INTERVAL-1):
            histogram[i]+=histogram[i+j+1]
        #This division is not really needed since
        #the slope will be the same anyway
        histogram[i] = histogram[i] / SMOOTH_INTERVAL
    return histogram

def calculate_threshold(histogram):
    """Returns an estimation of a good threshold"""
    maxVal, maxInd = 0, 0;
    #Find the maximum value and it's index
    for i in range(len(histogram)):
        if maxVal < histogram[i]:
            maxVal = histogram[i]
            maxInd = i
    startPoint = maxInd
    endPoint = maxInd + INTERVAL
    while endPoint < 256:
        #Calculate the slope
        deltaX = INTERVAL #endPoint - startPoint
        deltaY = histogram[endPoint] - histogram[startPoint]
        if deltaY == 0:
            k=deltaX
        else:
            k = deltaX / deltaY #the slope
        #print "K=",k
        if k >= K_MIN:
            return startPoint
        else:
            startPoint += INTERVAL
            endPoint += INTERVAL
    return -1 #We are not happy here
            
def threshold(image, threshold):
    """Apply a threshold to an image"""
    for i in range(len(image)):
        if image[i] > threshold:
            image[i] = ON_PIXEL #white
        else:
            image[i] = OFF_PIXEL #black
    return image


### OPERATIONS ON BINARY IMAGES ###

# This should perform some sanity checks to see that the
# search doesn't ascend or descend to another y-coordinate
def checkPerimeter(offset,image):
    """Check for ON pixels to the left and right"""
    left,right = 0,0
    index = offset
    #x = offset % SOBEL_IMAGE_WIDTH
    y,x = sobelIndexToYX(offset)
    for i in range(MAX_DISTANCE_TO_PIXEL):
        if image[index+i] == ON_PIXEL and x+i < SOBEL_IMAGE_WIDTH:
            right = 1
        if image[index-i] == ON_PIXEL and x-i >= 0:
            left = 1
    return right and left
        
def findNextOnPixel(offset, image, visited):
    """Find the next ON pixel in the image, starting at the offset"""
    for i in range(offset,len(image)):
        if image[i] == ON_PIXEL and i not in visited:
            return i
    return -1 #no more on pixels in the image

def findPotentialBlob(offset, image, visited):
    """If offset contains a possible blob return the new offset otherwise return -1"""
    blackStart, blackStop = -1, -1
    for i in range(1,MAX_DISTANCE_TO_PIXEL):
        index = offset+i*SOBEL_IMAGE_WIDTH
        if index >= SOBEL_IMAGE_SIZE:
            return -1
        pixel = image[index]
        if blackStart < 0 and pixel == OFF_PIXEL:
            blackStart = index
        elif blackStart >= 0 and blackStop < 0 and pixel == ON_PIXEL:
            blackStop = index-SOBEL_IMAGE_WIDTH #since we are on a white pixel
            half = blackStart + max((blackStop - blackStart)/2, SOBEL_IMAGE_WIDTH)
            if checkPerimeter(half,image) and half not in visited:
                #print ">>>>blackStart: ",blackStart
                #print "<<<<blackStop: ",blackStop
                return half
    return -1

def checkPixel(x, y, image):
    #offset = y * SOBEL_IMAGE_WIDTH + x
    #offset = SOBEL_LOOKUP_YX[y][x]
    return not(x<0 or y<0 or x>=SOBEL_IMAGE_WIDTH or y>=SOBEL_IMAGE_HEIGHT or
               image[sobelYXToIndex(y,x)]==ON_PIXEL or
               image[sobelYXToIndex(y,x)]==BLOB_PIXEL)
    
def floodFill(offset,image):
    blob = []
    head = [offset]
    while len(head) > 0:
	    offset = head.pop()
	    
            if len(blob) > MAX_BLOB_SIZE: # Rollback action
                for i in range(len(blob)):
                    image[blob.pop()] = OFF_PIXEL
                return []

            blob.append(offset)
            #image[offset] = ON_PIXEL
            image[offset] = BLOB_PIXEL
            #y,x = divmod(offset,SOBEL_IMAGE_WIDTH)
            y,x = sobelIndexToYX(offset)
            if checkPixel(x-1,y,image):
		head.append(offset - 1)
	    if checkPixel(x,y-1,image):
		head.append(offset - SOBEL_IMAGE_WIDTH)
	    if checkPixel(x+1,y,image):
		head.append(offset + 1)
	    if checkPixel(x,y+1,image):
		head.append(offset + SOBEL_IMAGE_WIDTH)
    return blob

def findAllBlobs(image):
    blob = []
    visited = []
    onPixel = findNextOnPixel(0,image,visited)
    while onPixel >= 0:
        regionGrowingOffset = findPotentialBlob(onPixel, image, visited)
        if regionGrowingOffset > -1:
            #print "Attempting region growing at: ", regionGrowingOffset
            visited.append(regionGrowingOffset)
            tmpBlob = floodFill(regionGrowingOffset,image)
            if len(tmpBlob) > MIN_BLOB_SIZE:
                blob.append(tmpBlob)
                visited += tmpBlob
        onPixel = findNextOnPixel(onPixel+1,image,visited)
    return blob

### OPERATIONS ON BLOBS ###
def calcBlobCenter(blob):
    blobcenter = []
    for i in range(len(blob)):
        tmp = blob[i]
        ymax,shit = sobelIndexToYX(max(tmp))
        ymin,shit = sobelIndexToYX(min(tmp))
        ycenter = ymin + (ymax - ymin)/2
        xmin = 1000
        xmax = 0
        for j in range(len(blob[i])):
            y,x = sobelIndexToYX(tmp[j])
            if x > xmax:
                xmax = x
            if x < xmin:
                xmin = x
        xcenter = xmin + (xmax - xmin)/2
        blobcenter.append((ycenter,xcenter))
    return blobcenter

def filterBlob(blobcenter, image):
    for i in range(len(blobcenter)):
        y,x = blobcenter[i]
        offset = sobelYXToIndex(y,x)
        if image[offset-SOBEL_IMAGE_WIDTH] == ON_PIXEL:
            blobcenter.pop(i)
            return blobcenter
        elif image[offset+SOBEL_IMAGE_WIDTH] == ON_PIXEL:
            blobcenter.pop(i)
            return blobcenter
        elif x+1 < SOBEL_IMAGE_WIDTH and image[offset+1] == ON_PIXEL: 
            blobcenter.pop(i)
            return blobcenter
        elif x-1 >= 0 and image[offset-1] == ON_PIXEL:
            blobcenter.pop(i)            
            return blobcenter
    return blobcenter

def calcBlobDistance(blobcenter):
    bloblen = len(blobcenter)
    blobdistance = [[0 for i in range(bloblen)] for i in range(bloblen)]
    for i in range(bloblen):
        for j in range(i,bloblen):
            y1,x1 = blobcenter[i]
            y2,x2 = blobcenter[j]
            dist = ((x1-x2)**2 + (y1-y2)**2)**0.5
            blobdistance[i][j] = dist
    return blobdistance

def fac(n):
    result = 1
    while n > 0:
        result = result * n
        n = n - 1
    return result

def combination(n,k):
    return fac(n)/(fac(k)*fac(n-k))        

def all3Combinations(array):
    tuple = []
    arraylen = len(array)
    i = 0
    j = 1
    k = 2
    for cntr in range(combination(arraylen,3)):
        ind1 = array[i]
        ind2 = array[j]
        ind3 = array[k]
        tuple.append([ind1,ind2,ind3])
        
        k += 1
        if k >= arraylen:
            j += 1
            k = j + 1
            if j + 1 >= arraylen:
                i += 1
                j = i + 1
                k = j + 1
                if i + 2 >= arraylen:
                    return tuple

def calcDistance((y1,x1),(y2,x2)):
    return ((x1-x2)**2 + (y1-y2)**2)**0.5

def checkDistances(dist1,dist2,dist3):
    """Check if all distances are within average +- 2"""
    avg = (dist1+dist2+dist3) / 3.0
    if dist1 > avg + DISTANCE_TOLERANCE or dist1 < avg - DISTANCE_TOLERANCE:
        return 0
    if dist2 > avg + DISTANCE_TOLERANCE or dist2 < avg - DISTANCE_TOLERANCE:
        return 0
    if dist3 > avg + DISTANCE_TOLERANCE or dist3 < avg - DISTANCE_TOLERANCE:
        return 0
    return 1

def calcAngle((uy,ux),(vy,vx)):
    dotprod = ux * vx + uy * vy
    ulen = (ux*ux + uy*uy)**0.5
    vlen = (vx*vx + vy*vy)**0.5
    return dotprod/(ulen*vlen)
    
def calcVectors((y1,x1),(y2,x2),(y3,x3)):
    uy,ux = (float(y2-y1),float(x2-x1))
    vy,vx = (float(y3-y1),float(x3-x1))
    return (uy,ux),(vy,vx)

# Should I consider the sum of the angles = 180
# instead of comparing them seperately to 60?
def checkAngles(angle1,angle2,angle3):
    """If angles are within interval return true"""
    # The reason for the inverted comparisment is
    # Due to the fact that cos(70) < cos(50)
    if angle1 < MAX_ANGLE or angle1 > MIN_ANGLE:
        return 0
    if angle2 < MAX_ANGLE or angle2 > MIN_ANGLE:
        return 0
    if angle3 < MAX_ANGLE or angle3 > MIN_ANGLE:
        return 0
    return 1

def calcDistanceToBall(dist):
    """Return the distance from the camera to the ball in cm"""
    return dist * SCALE_TO_REAL_WORLD

def calcCenterOfBall((y1,x1),(y2,x2),(y3,x3)):
    minx = min(x1,x2,x3)
    maxx = max(x1,x2,x3)
    miny = min(y1,y2,y3)
    maxy = max(y1,y2,y3)
    
    cx = minx + (maxx-minx)/2
    cy = miny + (maxy-miny)/2
    return cy,cx
        
    
def detectBall(blobcenter):
    numBlobs = len(blobcenter)
    combinations = all3Combinations(blobcenter)
    for i in range(len(combinations)):
        triple = combinations[i]
        dist1 = calcDistance(triple[0],triple[1])
        dist2 = calcDistance(triple[0],triple[2])
        dist3 = calcDistance(triple[1],triple[2])
        if checkDistances(dist1,dist2,dist3):
            (uy,ux),(vy,vx) = calcVectors(triple[0],triple[1],triple[2])
            angle1 = calcAngle((uy,ux),(vy,vx))
            (uy,ux),(vy,vx) = calcVectors(triple[1],triple[0],triple[2])
            angle2 = calcAngle((uy,ux),(vy,vx))
            (uy,ux),(vy,vx) = calcVectors(triple[2],triple[0],triple[1])
            angle3 = calcAngle((uy,ux),(vy,vx))
            #print "Angles: ",angle1," ",angle2," ",angle3
            if checkAngles(angle1,angle2,angle3):
                bdist = calcDistanceToBall(max(dist1,dist2,dist3))
                cy,cx = calcCenterOfBall(triple[0],triple[1],triple[2])
                return [(cy,cx),bdist]
    return []
                
            
            
### DEBUGGING METHODS ###    

def markBallCenter(coord,image):
    print repr(coord)
    cy,cx = coord
    index = sobelYXToIndex(cy,cx)
    image[index] = BLOB_PIXEL
    
def markCenter(blobcenter, image):
    for i in range(len(blobcenter)):
        y,x = blobcenter[i]
        index = sobelYXToIndex(y,x)
        image[index] = OFF_PIXEL

def paintBlobs(blob,image,color):
    for i in range(len(blob)):
        for j in range(len(blob[i])):
            image[blob[i][j]] = color

def profileIt():
    profile.run('realshit()','edgeprof')
    p=pstats.Stats('edgeprof')
    p.sort_stats('cumulative').print_stats(10)

def gnuplotform(histogram):
    for i in range(len(histogram)):
        print i, histogram[i]

def gnuplotform(histogram, filename):
    f = open(filename,'w')
    for i in range(len(histogram)):
        f.write(repr(i))
        f.write(' ')
        f.write(repr(histogram[i]))
        f.write('\n')


def testing():
    print "Opening image"
    im = Image.open("testIm2.bmp").convert("L")
    print "Obtaining image data"
    newIm = im.getdata()
    newIm2 = []
    for i in range(len(newIm)):
        newIm2.append(newIm[i])
    print "Finding all blobs"
    blob = findAllBlobs(newIm2)
    print "Found ", len(blob), " blobs"
    bla = Image.new("L",(SOBEL_IMAGE_WIDTH, SOBEL_IMAGE_HEIGHT))
    bla.putdata(newIm2)
    bla.show()

def realshit():
    print "Opening image"
    im = Image.open("../../images/images_real_ball/r1.bmp").convert("L")
    print "Obtaining image data"
    newIm = im.getdata()
    print "Detecting edges"
    newIm = fastSobel(newIm)
    print "Calculating histogram"
    hgram = histogram(newIm)
    gnuplotform(hgram, "raw_histogram.data")
    hgram = smooth_histogram(hgram)
    gnuplotform(hgram, "smooth_histogram.data")
    print "Determining a suitable threshold"
    threshold1 = calculate_threshold(hgram)
    print "Applying threshold at: ",threshold1
    newIm = threshold(newIm,threshold1)
    print "Finding all blobs"
    blob = findAllBlobs(newIm)
    print "Found ", len(blob), " blobs"
    if len(blob) < 3:
        print "Found to few blobs -> giving up!"
        return 0
    #for i in range(len(blob)):
    #    print len(blob[i])
    print "Calculating center for all blobs"
    blobcenter = calcBlobCenter(blob)
    #print "Filtering out blobs with centers too close to a border"
    #blobcenter = filterBlob(blobcenter,newIm)
    #print "Calculating distance between blobs"
    #blobdistance = calcBlobDistance(blobcenter)
    print "Determine position of ball(if any)"
    ballinfo = detectBall(blobcenter)
    if len(ballinfo) > 0:
        print "Found ball at: ", ballinfo[0]
        print "at a distance of: ", ballinfo[1]
        markBallCenter(ballinfo[0],newIm)
    else:
        print "Couldn't find a ball in the image"
    #print repr(blobcenter)
    #for i in range(len(blobdistance)):
    #    print repr(blobdistance[i])
    #paintBlobs(blob,newIm,BLOB_PIXEL)
    markCenter(blobcenter,newIm)
    bla = Image.new("L",(SOBEL_IMAGE_WIDTH, SOBEL_IMAGE_HEIGHT))
    bla.putdata(newIm)
    bla.show()

#testing()
realshit()
#profileIt()

